! @file test_compact_mask_index.f90
! @brief build and test use of mask index data structure
!
! Copyright  (C)  2012  Thomas Jahns <jahns@dkrz.de>
!
! @version 1.0
! @author Thomas Jahns <jahns@dkrz.de>
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
PROGRAM test_compact_mask_index
  USE ppm_base, ONLY: ppm_default_comm, assertion
  USE ppm_random, ONLY: irandr, irand, initialize_irand, a_randp_mt, frandp
  USE ppm_compact_mask_index, ONLY: range_compact_2d, range_compact_3d, &
       index_from_mask, index_from_mask_mt
  USE ppm_combinatorics, ONLY: permute
  USE ppm_extents, ONLY: extent, extent_start, extent_end, iinterval
  USE random_data, ONLY: random_rectilinear
#ifdef _OPENMP
  USE omp_lib, ONLY: omp_get_thread_num
#endif
  IMPLICIT NONE
  TYPE(range_compact_2d), ALLOCATABLE :: idx2d(:)
  TYPE(range_compact_3d), ALLOCATABLE :: idx3d(:)

  INTEGER :: i, j, k, n, r1, sseq(3), indices(3), rand_seed, tid
  INTEGER, PARAMETER :: max_iterations = 1000
  TYPE(iinterval), PARAMETER :: mask_size_max=iinterval(0, 40000000)
  TYPE(extent) :: mask_shape(3)
  INTEGER :: allocation_error, mask_size_lim
  LOGICAL, TARGET, ALLOCATABLE :: mask(:, :, :)
  LOGICAL :: p
  LOGICAL, POINTER :: mask_2d(:, :)
  REAL, TARGET, ALLOCATABLE :: accum(:, :, :), accum_ref(:, :, :)
  REAL, POINTER :: accum_2d(:, :), accum_ref_2d(:, :)
  REAL :: mask_fraction

!$omp parallel private(tid, i, j, k, n, p, rand_seed, mask_size_lim, &
!$omp        indices, allocation_error, mask_2d, accum_2d, accum_ref_2d) &
!$omp shared(accum, accum_ref, idx2d, idx3d, mask, mask_fraction, &
!$omp        mask_shape, ppm_default_comm, sseq)
#ifdef _OPENMP
  tid = omp_get_thread_num()
#else
  tid = 0
#endif
  CALL initialize_irand(ppm_default_comm, 0, rand_seed)
  PRINT '(2(a,i0))', 'thread id=', tid, ', random seed=', rand_seed
!$omp master
  n = 0
  allocation_loop: DO WHILE(.NOT. ALLOCATED(mask) .AND. n <max_iterations)
    mask_size_lim = irandr(mask_size_max)
    CALL random_rectilinear(mask_shape, mask_size_lim)
    ALLOCATE(mask(extent_start(mask_shape(1)):extent_end(mask_shape(1)), &
         extent_start(mask_shape(2)):extent_end(mask_shape(2)), &
         extent_start(mask_shape(3)):extent_end(mask_shape(3))), &
         accum(extent_start(mask_shape(1)):extent_end(mask_shape(1)), &
         extent_start(mask_shape(2)):extent_end(mask_shape(2)), &
         extent_start(mask_shape(3)):extent_end(mask_shape(3))), &
         accum_ref(extent_start(mask_shape(1)):extent_end(mask_shape(1)), &
         extent_start(mask_shape(2)):extent_end(mask_shape(2)), &
         extent_start(mask_shape(3)):extent_end(mask_shape(3))), &
         stat=allocation_error)
    n = n + 1
  END DO allocation_loop
  IF (.NOT. ALLOCATED(mask)) &
       WRITE (0, '(a,i0)') 'allocation_error=', allocation_error
  CALL assertion(ALLOCATED(mask), &
       __FILE__, &
       __LINE__, &
       'failed to generate allocatable mask')
!$omp end master
!$omp barrier
  CALL a_randp_mt(accum)
!$omp master
  ! ensure mask is densely-populated
  mask_fraction = 0.25 * frandp()
!$omp end master
!$omp barrier
!$omp workshare
  accum_ref = accum
  mask = accum >= mask_fraction
!$omp end workshare

  ! 2D
  mask_2d => mask(:, :, LBOUND(mask, 3))
  accum_2d => accum(:, :, LBOUND(accum, 3))
  accum_ref_2d => accum_ref(:, :, LBOUND(accum_ref, 3))
!$omp single
  sseq(1:2) = MERGE((/ 1, 2 /), (/ 2, 1 /), irand()>0)
  CALL index_from_mask(idx2d, mask_2d, offsets=LBOUND(mask_2d), &
       sseq=sseq(1:2))
!$omp end single

! older intel compilers cannot parallelize the following where correctly
#if !_OPENMP || !defined __INTEL_COMPILER || __INTEL_COMPILER > 1200
!$omp workshare
  WHERE(mask_2d) accum_ref_2d(:, :) = accum_ref_2d(:, :) + 1.0
!$omp end workshare
#else
  ! since older intel compilers cannot parallelize the above where correctly,
  ! use an explicit DO
!$omp do
  DO j = LBOUND(mask_2d, 2), UBOUND(mask_2d, 2)
    DO i = LBOUND(mask_2d, 1), UBOUND(mask_2d, 1)
      IF (mask_2d(i, j)) THEN
        accum_ref_2d(i, j) = accum_ref_2d(i, j) + 1.0
      END IF
    END DO
  END DO
#endif

  n = SIZE(idx2d)
  IF (sseq(1) == 1) THEN
!$omp do
    DO r1 = 1, n
      j = idx2d(r1)%tl_ss
      DO i = idx2d(r1)%ll_range%first, idx2d(r1)%ll_range%last
        accum_2d(i, j) = accum_2d(i, j) + 1.0
      END DO
    END DO
!$omp end do
  ELSE ! sseq(1) == 2
!$omp do
    DO r1 = 1, n
      i = idx2d(r1)%tl_ss
      DO j = idx2d(r1)%ll_range%first, idx2d(r1)%ll_range%last
        accum_2d(i, j) = accum_2d(i, j) + 1.0
      END DO
    END DO
!$omp end do
  END IF
  p = .TRUE.
  n = UBOUND(accum_2d, 2)
!$omp do
  DO j = LBOUND(accum_2d, 2), n
    p = p .AND. ALL(accum_2d(:, j) == accum_ref_2d(:, j))
  END DO
!$omp end do nowait
  CALL assertion(p, &
       __FILE__, &
       __LINE__, &
       'result of WHERE(mask) and index iteration differs')

!$omp end parallel
  ! 3D
  sseq = (/ 1, 2, 3 /)
  CALL permute(sseq)
  CALL index_from_mask_mt(idx3d, mask(:,:,:), offsets=LBOUND(mask), sseq=sseq)
!$omp parallel  private(tid, i, j, k, n, p, rand_seed, &
!$omp        indices, allocation_error) firstprivate(sseq) &
!$omp shared(accum, accum_ref, idx2d, idx3d, mask, mask_fraction, &
!$omp        mask_shape, ppm_default_comm)

  ! older intel compilers cannot parallelize the following where correctly
#if !_OPENMP || !defined __INTEL_COMPILER || __INTEL_COMPILER > 1200
!$omp workshare
  WHERE(mask(:, :, :)) accum_ref(:, :, :) = accum_ref(:, :, :) + 1.0
!$omp end workshare
#else
  ! since older intel compilers cannot parallelize the above where correctly,
  ! use an explicit DO
!$omp do
  DO k = LBOUND(mask, 3), UBOUND(mask, 3)
    DO j = LBOUND(mask, 2), UBOUND(mask, 2)
      DO i = LBOUND(mask, 1), UBOUND(mask, 1)
        IF (mask(i, j, k)) THEN
          accum_ref(i, j, k) = accum_ref(i, j, k) + 1.0
        END IF
      END DO
    END DO
  END DO
#endif

  n = SIZE(idx3d)
!$omp do
  DO r1 = 1, n
    j = idx3d(r1)%tl_ss(1)
    k = idx3d(r1)%tl_ss(2)
    DO i = idx3d(r1)%ll_range%first, idx3d(r1)%ll_range%last
      ! do not allocate temporary
      indices(sseq(1)) = i
      indices(sseq(2)) = j
      indices(sseq(3)) = k
      accum(indices(1), indices(2), indices(3)) &
           = accum(indices(1), indices(2), indices(3)) + 1.0
    END DO
  END DO
!$omp end do
  p = .TRUE.
  n = UBOUND(accum, 3)
!$omp do
  DO k = LBOUND(accum, 3), n
    p = p .AND. ALL(accum(:, :, k) == accum_ref(:, :, k))
  END DO
!$omp end do nowait
  CALL assertion(p, &
       __FILE__, &
       __LINE__, &
       'result of WHERE(mask) and index iteration differs')
!$omp end parallel
END PROGRAM test_compact_mask_index
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-markup: "doxygen"
! license-default: "bsd"
! End:
!
