! test_sparse_mask_index.f90 --- test sparse mask index
!
! Copyright  (C)  2011  Thomas Jahns <Thomas.Jahns@gmx.net>
!
! Version: 1.0
! Keywords: sparse mask index test
! Author: Thomas Jahns <Thomas.Jahns@gmx.net>
! Maintainer: Thomas Jahns <Thomas.Jahns@gmx.net>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
PROGRAM test_sparse_mask_index
  USE ppm_base, ONLY: ppm_default_comm, assertion
  USE ppm_combinatorics, ONLY: permute
  USE ppm_extents, ONLY: extent, extent_start, extent_end, iinterval
  USE ppm_random, ONLY: irandr, lrand, initialize_irand, a_randp_mt, frandp
  USE ppm_sparse_mask_index, ONLY: index_sparse_nd, index_from_mask
  USE random_data, ONLY: random_rectilinear
#ifdef _OPENMP
  USE omp_lib, ONLY: omp_get_thread_num
#endif
  IMPLICIT NONE
  TYPE(index_sparse_nd) :: idx

  INTEGER :: i, j, k, n, r1, r2, r3, iblock, jblock, sseq(3), indices(3)
  INTEGER, PARAMETER :: max_iterations = 1000
  TYPE(iinterval), PARAMETER :: mask_size_range=iinterval(0, 400000)
  TYPE(extent) :: mask_dims(3)
  INTEGER :: mask_size_ub
  INTEGER :: allocation_error, rand_seed, tid
  LOGICAL, TARGET, ALLOCATABLE :: mask(:, :,:)
  LOGICAL, POINTER :: mask_2d(:, :), mask_1d(:)
  LOGICAL :: p
  REAL, TARGET, ALLOCATABLE :: accum(:, :, :), accum_ref(:, :, :)
  REAL, POINTER :: accum_2d(:, :), accum_ref_2d(:, :), &
       accum_1d(:), accum_ref_1d(:)
  REAL :: mask_fraction

!$omp parallel &
!$omp private(i, tid, rand_seed, n, mask_size_ub, allocation_error, r2, r3, &
!$omp iblock, jblock, k, indices, mask_2d, accum_2d, accum_ref_2d, &
!$omp mask_1d, accum_1d, accum_ref_1d) &
!$omp & shared(accum, accum_ref, mask, ppm_default_comm, mask_dims, &
!$omp & mask_fraction, idx, p, sseq)
#ifdef _OPENMP
  tid = omp_get_thread_num()
#else
  tid = 0
#endif
  CALL initialize_irand(ppm_default_comm, 0, rand_seed)
  PRINT '(2(a,i0))', 'thread id=', tid, ', random seed=', rand_seed
!$omp master
  n = 0
  allocation_loop: DO WHILE(.NOT. ALLOCATED(mask) .AND. n < max_iterations)
    mask_size_ub = irandr(mask_size_range)
    CALL random_rectilinear(mask_dims, mask_size_ub)
    ALLOCATE(mask(extent_start(mask_dims(1)):extent_end(mask_dims(1)), &
         &        extent_start(mask_dims(2)):extent_end(mask_dims(2)), &
         &        extent_start(mask_dims(3)):extent_end(mask_dims(3))), &
         accum(extent_start(mask_dims(1)):extent_end(mask_dims(1)), &
         &     extent_start(mask_dims(2)):extent_end(mask_dims(2)), &
         &     extent_start(mask_dims(3)):extent_end(mask_dims(3))), &
         accum_ref(extent_start(mask_dims(1)):extent_end(mask_dims(1)), &
         &         extent_start(mask_dims(2)):extent_end(mask_dims(2)), &
         &         extent_start(mask_dims(3)):extent_end(mask_dims(3))), &
         stat=allocation_error)
    n = n + 1
  END DO allocation_loop
  IF (.NOT. ALLOCATED(mask)) &
       WRITE (0, '(a,i0)') 'allocation_error=', allocation_error
  CALL assertion(ALLOCATED(mask), &
       __FILE__, &
       __LINE__, 'failed to generate allocatable mask')
  ! the following is redundant, but needed for compiler analysis of allocation
  ! status
  IF (.NOT. ALLOCATED(accum_ref)) THEN
    STOP 1
  END IF
!$omp end master
!$omp barrier
  CALL a_randp_mt(accum)
!$omp master
  ! ensure mask is sparse
  mask_fraction = 0.75 + 0.25 * frandp()
!$omp end master
!$omp barrier
!$omp workshare
  accum_ref = accum
  mask = accum >= mask_fraction
!$omp end workshare

  ! 1D-case
  mask_1d => mask(:, LBOUND(mask, 2), LBOUND(mask, 3))
  accum_1d => accum(:, LBOUND(accum, 2), LBOUND(accum, 3))
  accum_ref_1d => accum_ref(:, LBOUND(accum_ref, 2), LBOUND(accum_ref, 3))
!$omp single
  idx = index_from_mask(mask_1d, offsets=LBOUND(mask_1d))
!$omp end single
!$omp workshare
  WHERE(mask_1d) accum_ref_1d = accum_ref_1d + 1.0
!$omp end workshare
  n = idx%tl_ranges
!$omp do
  DO r1 = 1, n
    DO i = idx%ranges(r1)%first, idx%ranges(r1)%last
      accum_1d(i) = accum_1d(i) + 1.0
    END DO
  END DO
!$omp end do
  p = .TRUE.
  n = UBOUND(accum_1d, 1)
!$omp do
  DO i = LBOUND(accum_1d, 1), n
    p = p .AND. accum_1d(i) == accum_ref_1d(i)
  END DO
!$omp end do
  CALL assertion(p, &
       __FILE__, &
       __LINE__, &
       'result of WHERE(mask) and index iteration differs')
  ! 2D
  mask_2d => mask(:, :, LBOUND(mask, 3))
  accum_2d => accum(:, :, LBOUND(accum, 3))
  accum_ref_2d => accum_ref(:, :, LBOUND(accum_ref, 3))
!$omp single
  sseq(1:2) = MERGE((/ 1, 2 /), (/ 2, 1 /), lrand())
  idx = index_from_mask(mask_2d, sseq=sseq(1:2), offsets=LBOUND(mask_2d))
!$omp end single
!$omp workshare
  WHERE(mask_2d) accum_ref_2d = accum_ref_2d + 1.0
!$omp end workshare
  n = 2 * idx%tl_ranges
  IF (sseq(1) == 1) THEN
!$omp do
    DO r1 = 1, n, 2
      DO j = idx%ranges(r1)%first, idx%ranges(r1)%last
        r2 = idx%ranges(r1+1)%first + j - idx%ranges(r1)%first
        DO iblock = idx%ranges(r2)%first, idx%ranges(r2)%last
          DO i = idx%ranges(iblock)%first, idx%ranges(iblock)%last
            accum_2d(i, j) = accum_2d(i, j) + 1.0
          END DO
        END DO
      END DO
    END DO
!$omp end do
  ELSE ! sseq(1) == 2
!$omp do
    DO r1 = 1, n, 2
      DO i = idx%ranges(r1)%first, idx%ranges(r1)%last
        r2 = idx%ranges(r1+1)%first + i - idx%ranges(r1)%first
        DO iblock = idx%ranges(r2)%first, idx%ranges(r2)%last
          DO j = idx%ranges(iblock)%first, idx%ranges(iblock)%last
            accum_2d(i, j) = accum_2d(i, j) + 1.0
          END DO
        END DO
      END DO
    END DO
!$omp end do
  END IF
  p = .TRUE.
  n = UBOUND(accum_2d, 2)
!$omp do
  DO j = LBOUND(accum_2d, 2), n
    p = p .AND. ALL(accum_2d(:, j) == accum_ref_2d(:, j))
  END DO
!$omp end do
  CALL assertion(p, &
       __FILE__, &
       __LINE__, &
       'result of WHERE(mask) and index iteration differs')
  ! 3D
!$omp single
  sseq = (/ 1, 2, 3 /)
  CALL permute(sseq)
  idx = index_from_mask(mask, sseq=sseq, offsets=LBOUND(mask))
!$omp end single
!$omp workshare
  WHERE(mask(:, :, :)) accum_ref(:, :, :) = accum_ref(:, :, :) + 1.0
!$omp end workshare
  n = 2 * idx%tl_ranges
!$omp do
  DO r1 = 1, n, 2
    DO k = idx%ranges(r1)%first, idx%ranges(r1)%last
      r2 = idx%ranges(r1+1)%first + k - idx%ranges(r1)%first
      DO jblock = idx%ranges(r2)%first, idx%ranges(r2)%last, 2
        DO j = idx%ranges(jblock)%first, idx%ranges(jblock)%last
          r3 = idx%ranges(jblock+1)%first + j - idx%ranges(jblock)%first
          DO iblock = idx%ranges(r3)%first, idx%ranges(r3)%last
            DO i = idx%ranges(iblock)%first, idx%ranges(iblock)%last
              indices(sseq) = (/ i, j, k /)
              accum(indices(1), indices(2), indices(3)) &
                   = accum(indices(1), indices(2), indices(3)) + 1.0
            END DO
          END DO
        END DO
      END DO
    END DO
  END DO
!$omp end do
  p = .TRUE.
  n = UBOUND(accum, 3)
!$omp do reduction(.and. : p)
  DO k = LBOUND(accum, 3), n
    p = p .AND. ALL(accum(:, :, k) == accum_ref(:, :, k))
  END DO
!$omp end do
  CALL assertion(p, &
       __FILE__, &
       __LINE__, &
       'result of WHERE(mask) and index iteration differs')
!$omp end parallel
END PROGRAM test_sparse_mask_index
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-default: "bsd"
! End:
